'''
usage:
  bf cache [options] clear
  bf cache [options] compact

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt

def main():
    args = docopt(__doc__)

    from blackfynn.cache import cache

    if args['clear']:
        print "Clearing cache...",
        cache.clear()
        print "done."
    elif args['compact']:
        cache.start_compaction(async=False)



'''
usage:
  bf path [options] <item>

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
from cli_utils import print_path_tree, get_client
        
def main():
    args = docopt(__doc__)

    bf = get_client()
    
    item = bf.get(args['<item>'])
    print_path_tree(bf, [item])



'''usage:
  bf [options] [<command>] [<args>...]

Available commands:
  append        Append data to a datset
  cache         Perform cache operations
  create        Create dataset or collection
  dataset       Dataset actions
  delete        Delete item
  env           Display bf environment
  path          Show path to item
  profile       Profile management
  upload        Upload file(s) or directory
  search        Search datasets
  show          Show datasets, orgs, or contents of an item

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
import os

import blackfynn
from blackfynn import Blackfynn, settings

def blackfynn_cli():
    args = docopt(__doc__,
                  version='bf version {}'.format(blackfynn.__version__),
                  options_first=True)

    #Display warning message if config.ini is not found
    if args['<command>'] != 'profile':
        if not os.path.exists(settings.config_file):
            print("\033[31m* Warning: No config file found, run 'bf profile' to start the setup assistant\033[0m")

    #Try to use profile specified by --profile, exit if invalid
    try:
        if args['--profile'] is not None:
            settings.use_profile(args['--profile'])
    except Exception, e:
        exit(e)
        
    if args['<command>'] == 'env':
        import bf_env
        bf_env.main()
    elif args['<command>'] == 'cache':
        import bf_cache
        bf_cache.main()
    elif args['<command>'] == 'create':
        import bf_create
        bf_create.main()
    elif args['<command>'] == 'delete':
        import bf_delete
        bf_delete.main()
    elif args['<command>'] == 'show':
        import bf_show
        bf_show.main()
    elif args['<command>'] == 'path':
        import bf_path
        bf_path.main()        
    elif args['<command>'] == 'upload':
        import bf_upload
        bf_upload.main()
    elif args['<command>'] == 'append':
        import bf_append
        bf_append.main()
    elif args['<command>'] == 'search':
        import bf_search
        bf_search.main()
    elif args['<command>'] == 'dataset':
        import bf_dataset
        bf_dataset.main()
    elif args['<command>'] == 'profile':
        import bf_profile
        bf_profile.main()
    elif args['<command>'] in ['help',None]:
        print(__doc__.strip('\n'))
        return
    else:
        exit("Invalid command: '{}'\nSee 'bf help' for available commands".format(args['<command>']))



'''Show an item, list datasets, or orgs

usage:
  bf show [options] (datasets | ds)
  bf show [options] orgs
  bf show [options] <item>

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt

from cli_utils import get_client

def main():
    args = docopt(__doc__)

    bf = get_client()

    if args['datasets'] or args['ds']:
        print "Datasets:"
        for ds in bf.datasets():
            print " * {} (id: {})".format(ds.name, ds.id)
    elif args['orgs']:
        print('Organizations:')
        for o in bf.organizations():
            print " * {} (id: {})".format(o.name, o.id)
    elif args['<item>'][:2] == 'N:':
        try:
            item = bf.get(args['<item>'])
            print item
            if hasattr(item, 'items'):
                print "CONTENTS:"
                for i in item.items:
                    print " * {}".format(i)
            if hasattr(item, 'channels'):
                print "CHANNELS:"
                for ch in item.channels:
                    print " * {} (id: {})".format(ch.name, ch.id)
        except:
            print("Invalid item: '{}'".format(args['<item>']))




__title__ = 'blackfynn'
__version__ = '1.8.4'

from .config import settings

# main client
from .client import Blackfynn

# base models
from .models import (
    BaseNode,
    Property,
    Organization,
    File,
    DataPackage,
    Collection,
    Dataset,
    Tabular,
    TabularSchema,
    TimeSeries,
    TimeSeriesChannel,
    TimeSeriesAnnotation,
    LedgerEntry
)



# -*- coding: utf-8 -*-

from blackfynn.api.base import APIBase
from blackfynn.models import LedgerEntry

class LedgerAPI(APIBase):
    base_uri = "/ledger"
    name = 'ledger'

    def create_ledger_entry(self, entry):
        return self._post( self._uri('/entries'), data=entry.as_dict())

    def get_ledger_entries(self, orgId, start, end):
        s = start.replace(microsecond=0).isoformat() + 'Z' #the 'Z' is because python's datetime doesn't conform to the ISO 8601 standard
        e = end.replace(microsecond=0).isoformat()  + 'Z'
        uri = self._uri('/entries?orgId={orgId}&start={start}&end={end}', orgId=orgId, start=s,end=e)
        return [ LedgerEntry.from_dict( r ) for r in self._get( uri ) ]




def get_item_path(bf, item):
    if hasattr(item, 'parent'):
        parent = bf.get(item.parent)
        return get_item_path(bf, parent) + [item]
    else:
        return [item]

def make_tree(path):
    if len(path)==1:
        value = path[0]
    else:
        value = make_tree(path[1:])
    return {path[0].id: value}

def merge(a, b, path=None):
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a

def print_tree(tree, objects, indent=0):
    for key, value in tree.iteritems():
        is_dict = isinstance(value, dict)
        char = '-' if is_dict else '*'
        print ' '*indent+char, objects[key]
        if is_dict:
            print_tree(value, objects, indent=indent+2)

def print_path_tree(bf, results):
    paths = [get_item_path(bf, r) for r in results]
    path_objects = {o.id:o for p in paths for o in p}
    trees = [make_tree(path) for path in paths]
    print_tree(reduce(merge, trees), path_objects)

def print_collaborator_edit_resp(resp):
    for key, value in resp.iteritems():
        if value['success']:
            print " - {}: Success".format(key)
        else:
            print " - {}: Error - {}".format(key, value['message'])

def recursively_upload(bf,dest,files):
    import os
    from blackfynn import Collection
    
    dirs = [f for f in files if os.path.isdir(f)]
    files = [f for f in files if os.path.isfile(f)]
    if not dest.startswith('N:'):
        # maybe destination is dataset name?
        dest = bf.get_dataset(dest)

    if len(files) > 0:
        bf._api.io.upload_files(dest, files, display_progress=True)
    for d in dirs:
        name = os.path.basename(os.path.normpath(d))
        print 'Uploading to {}'.format(name)
        c = Collection(name)
        parent = bf.get(dest)
        parent.add(c)
        files = [os.path.join(d,f) for f in os.listdir(d) if not f.startswith('.')]
        recursively_upload(bf,c.id,files)

def get_client():
    from blackfynn import Blackfynn

    try:
        bf = Blackfynn()
        return bf
    except:
        exit("Unable to authenticate against Blackfynn using the specified API token.")



import os
import configparser
import sys

class Settings(object):
    def __init__(self):
        self.profiles = {}
        
        self._load_defaults()
        self._load_global()
        self._load_profiles()

        try:
            self.use_profile(self.config['global']['default_profile'])
        except:
            self.use_profile('global')

        self._load_eVars()
        
    def _load_defaults(self):
        # GET/MAKE REQUIRED DIRECTORIES
        #=============================================
        self.blackfynn_dir = os.environ.get('BLACKFYNN_LOCAL_DIR', '{}/.blackfynn'.format(os.path.expanduser('~')))
        if not os.path.exists(self.blackfynn_dir):
            os.makedirs(self.blackfynn_dir)
            
        self.cache_dir = os.environ.get('BLACKFYNN_CACHE_LOC', os.path.join(self.blackfynn_dir,'cache'))
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        # GET CONFIG OBJECT
        #============================================
        self.config_file = os.path.join(self.blackfynn_dir,'config.ini')
        self.config = configparser.ConfigParser()
        self.config.read(self.config_file)

            
        # DEFAULT SETTINGS
        #=============================================
        self.defaults = {    
            # blackfynn api locations
            'api_host'                    : 'https://api.blackfynn.io',
            'streaming_api_host'          : 'https://streaming.blackfynn.io',
            
            # blackfynn API token/secret
            'api_token'                   : None,
            'api_secret'                  : None,
            
            # streaming
            'stream_name'                 : 'prod-stream-blackfynn',
            'stream_aws_region'           : 'us-east-1',
            'stream_max_segment_size'     : 5000,
            
            # all requests
            'max_request_time'            : 120, # two minutes
            'max_request_timeout_retries' : 2,
            
            #io
            'max_upload_workers'          : 10,
            
            # timeseries
            'max_points_per_chunk'        : 10000,
            
            # s3 (amazon/local)
            's3_host'                     : '',
            's3_port'                     : '',
            
            # logging
            'log_level'                   : 'INFO',
            
            # cache
            'cache_index'                 : os.path.join(self.cache_dir, 'index.db'),
            'cache_max_size'              : 2048, #os.environ.get('BLACKFYNN_CACHE_MAX_SIZE', 2048) # in MBs
            'cache_inspect_interval'      : 1000, #os.environ.get('BLACKFYNN_CACHE_INSPECT_EVERY', 1000) # num page writes
            'ts_page_size'                : 3600, #int(os.environ.get('BLACKFYNN_TS_PAGE_SIZE', 3600)) # num points / page
            'use_cache'                   : True, #bool(int(os.environ.get('BLACKFYNN_USE_CACHE', True)))
        }

    def _load_global(self):
        self.default_profile = None
        self.profiles['global'] = self.defaults.copy()
        if 'global' in self.config:
            for key, value in self.config['global'].items():
                if value == 'none'            : self.profiles['global'][key] = None
                elif value.lower() == 'true'  : self.profiles['global'][key] = True
                elif value.lower() == 'false' : self.profiles['global'][key] = False
                elif value.isdigit()          : self.profiles['global'][key] = int(value)
                else                          : self.profiles['global'][key] = str(value)

    def _load_eVars(self):
        eVars_dict = {
            'api_host'               : 'BLACKFYNN_API_LOC',
            'streaming_api_host'     : 'BLACKFYNN_STREAMING_API_LOC',
            'api_token'              : 'BLACKFYNN_API_TOKEN',
            'api_secret'             : 'BLACKFYNN_API_SECRET',
            'stream_name'            : 'BLACKFYNN_STREAM_NAME',
            
            'cache_max_size'         : 'BLACKFYNN_CACHE_MAX_SIZE',
            'cache_inspect_interval' : 'BLACKFYNN_CACHE_INSPECT_EVERY',
            'ts_page_size'           : 'BLACKFYNN_TS_PAGE_SIZE',
            'use_cache'              : 'BLACKFYNN_USE_CACHE',
            'log_level'              : 'BLACKFYNN_LOG_LEVEL',
            'default_profile'        : 'BLACKFYNN_PROFILE',

            # advanced
            's3_host'                : 'S3_HOST',
            's3_port'                : 'S3_PORT'
        }

        self.eVars = {}
        for key, eVar in eVars_dict.items():
            value = os.environ.get(eVar,None)
            if value is not None:
                self.eVars[key] = [value,eVar]
        
        self.__dict__.update((key, os.environ.get(value,self.__dict__[key])) for key,value in eVars_dict.items())

        if 'default_profile' in self.eVars:
            try:
                self.use_profile(self.eVars['default_profile'][0])
            except:
                sys.exit("Invalid value '{}' for environment variable {}".format(self.eVars['default_profile'][0],self.eVars['default_profile'][1]))
                
    def _load_profiles(self):
        for name in self.config.sections():
            if name is not 'global':
                self.profiles[name] = self.profiles['global'].copy()
                for key, value in self.config[name].items():
                    if value == 'none'            : self.profiles[name][key] = None
                    elif value.lower() == 'true'  : self.profiles[name][key] = True
                    elif value.lower() == 'false' : self.profiles[name][key] = False
                    elif value.isdigit()          : self.profiles[name][key] = int(value)
                    else                          : self.profiles[name][key] = str(value)

    def use_profile(self,name):
        if name is not None:
            if name not in self.profiles:
                raise Exception('Invaid profile name')
            else:
                self.__dict__.update(self.profiles[name])
                self.active_profile = name
                if name is 'global': self.active_profile = 'none'

settings = Settings()



import os
import time
import sqlite3
import numpy as np
import pandas as pd
from glob import glob
import multiprocessing as mp
from itertools import groupby
from datetime import datetime

# blackfynn-specific
from blackfynn import settings
from blackfynn.utils import usecs_to_datetime, log
from blackfynn.models import DataPackage, TimeSeriesChannel

def filter_id(some_id):
    return some_id.replace(':','_').replace('-','_')

def remove_old_pages(cache, mbdiff):
    # taste the rainbow!
    n = int(1.5 * ((mbdiff * 1024*1024) / 100) / cache.page_size) + 5

    # 2. Delete some pages from cache
    with cache.index_con as con:
        log.debug("Cache - removing {} pages...".format(n))
        # find the oldest/least accessed pages
        q = """
            SELECT channel,page,access_count,last_access
            FROM ts_pages
            ORDER BY last_access ASC, access_count ASC
            LIMIT {num_pages}
        """.format(num_pages=n)
        pages = con.execute(q).fetchall()

    # remove the selected pages
    pages_by_channel = groupby(pages, lambda x: x[0])
    for channel, page_group in pages_by_channel:
        _,pages,counts,times = zip(*page_group)
        # remove page files
        cache.remove_pages(channel, *pages)

    with cache.index_con as con:
        con.execute("VACUUM")

    log.debug('Cache - {} pages removed.'.format(n))
    return n


def compact_cache(cache):
    log.debug('Inspecting cache...')
    wait = 2
    max_mb = settings.cache_max_size
    current_mb = (cache.size/(1024.0*1024))
    desired_mb = 0.9*max_mb 
    while current_mb > desired_mb:
        log.debug('Cache - current: {:02f} MB, maximum: {} MB'.format(current_mb, max_mb))
        try:
            remove_old_pages(cache, current_mb-desired_mb)
        except sqlite3.OperationalError:
            log.debug('Cache - Index DB was locked, waiting {} seconds...'.format(wait))
            if wait >= 1024:
                log.error('Cache - Unable to compact cache!')
                return # silently fail
            time.sleep(wait)
            wait = wait*2
        current_mb = (cache.size/(1024.0*1024))


class Cache(object):
    def __init__(self):
        self.dir           = settings.cache_dir
        self.index_loc     = settings.cache_index
        self.index_con     = sqlite3.connect(self.index_loc, timeout=60)
        self.write_counter = 0

        # this might be replaced with existing page size (from DB)
        self.page_size = settings.ts_page_size

        # setup DB
        self.init_tables()

        # check/compact cache if too big
        self.start_compaction()

    def init_tables(self):
        with self.index_con as con:
            self.init_index_table(con)
            self.init_settings_table(con)

    def init_index_table(self, con):
        # check for index table
        q = "SELECT name FROM sqlite_master WHERE type='table' AND name='ts_pages'"
        r = con.execute(q)
        if r.fetchone() is None:
            log.info('Cache - Creating \'ts_pages\' table')
            # create index table
            q = """
                CREATE TABLE ts_pages (
                    channel CHAR(50) NOT NULL,
                    page INTEGER NOT NULL,
                    access_count INTEGER NOT NULL,
                    last_access DATETIME NOT NULL,
                    has_data BOOLEAN,
                    PRIMARY KEY (channel, page))
            """
            con.execute(q)

    def init_settings_table(self, con):
        # check for settings table
        q = "SELECT name FROM sqlite_master WHERE type='table' AND name='settings'"
        r = con.execute(q)
        if r.fetchone() is None:
            log.info('Cache - Creating \'settings\' table')
            # create settings table
            q = """
                CREATE TABLE settings (
                    ts_page_size INTEGER NOT NULL,
                    max_bytes    INTEGER NOT NULL,
                    modified     DATETIME)
            """
            con.execute(q)

            # insert settings values
            q = """
                INSERT INTO settings
                VALUES ({page_size},{max_bytes},'{time}')
            """.format(
                page_size = self.page_size,
                max_bytes = settings.cache_max_size,
                time      = datetime.now().isoformat())
            con.execute(q)

        else:
            # settings table exists... get page size
            result = con.execute("SELECT ts_page_size FROM settings").fetchone()
            if result is not None:
                self.page_size = result[0]
                if settings.ts_page_size != self.page_size:
                    log.warn('Using existing page_size={} from DB settings (user specified page_size={})' \
                        .format( self.page_size, settings.ts_page_size))
            else:
                # somehow, there is no entry
                self.page_size = settings.ts_page_size

    def set_page(self, channel, page, has_data):
        with self.index_con as con:
            q = "INSERT INTO ts_pages VALUES ('{channel}',{page},0,'{time}',{has_data})".format(
                    channel=channel.id,
                    page=page,
                    time=datetime.now().isoformat(),
                    has_data=int(has_data))
            con.execute(q)

    def set_page_data(self, channel, page, data, update=False):
        has_data = False if data is None else len(data)>0
        if has_data:
            # there is data, write it to file
            filename = self.page_file(channel.id, page, make_dir=True)
            data.name = 'values'
            data.reset_index().to_feather(filename)
            self.page_written()
        try:
            if update:
                # modifying an existing page entry
                self.update_page(channel, page, has_data)
            else:
                # adding a new page entry
                self.set_page(channel, page, has_data)
        except sqlite3.IntegrityError:
            # page already exists - ignore
            pass

    def check_page(self, channel, page):
        """
        Does page exist in cache?
        """
        with self.index_con as con:
            q = """ SELECT page
                    FROM   ts_pages
                    WHERE  channel='{channel}' AND page={page}
            """.format(channel=channel.id, page=page)
            r = con.execute(q).fetchone()
            return r is not None

    def page_has_data(self, channel, page):
        with self.index_con as con:
            q = """
                SELECT has_data
                FROM   ts_pages
                WHERE  channel='{channel}' AND page={page}
            """.format(channel=channel.id, page=page)
            r = con.execute(q).fetchone()
            return None if r is None else bool(r[0])

    def get_page_data(self, channel, page):
        has_data = self.page_has_data(channel, page) 
        if has_data is None:
            # page not present in cache
            return None
        elif not has_data:
            # page is empty
            return pd.Series([])

        # page has data, let's get it
        filename = self.page_file(channel.id, page, make_dir=True)
        if os.path.exists(filename):
            # get page data from file
            series = pd.read_feather(filename).set_index('index')['values']
            # update access count
            self.update_page(channel, page, has_data)
            return series
        else:
            # page file has been deleted recently?
            log.warn('Page file not found: {}'.format(filename))
            return None

    def update_page(self, channel, page, has_data=True):
       with self.index_con as con:
            q = """
                UPDATE ts_pages 
                SET access_count = access_count + 1,
                    last_access  = '{now}',
                    has_data     = {has_data}
                WHERE channel='{channel}' AND page='{page}'
            """.format(channel=channel.id,
                       page=page,
                       has_data=int(has_data),
                       now=datetime.now().isoformat())
            con.execute(q) 

    def page_written(self):
        # cache compaction?
        self.write_counter += 1
        if self.write_counter > settings.cache_inspect_interval:
            self.write_counter = 0
            self.start_compaction()

    def start_compaction(self, async=True):
        if async:
            # spawn cache compact job
            p = mp.Process(target=compact_cache, args=(self,))
            p.start() 
        else:
            log.info('Compacting cache...')
            compact_cache(self)
            log.info('Done')

    def remove_pages(self, channel_id, *pages):
        # remove page data files
        for page in pages:
            filename = self.page_file(channel_id,page)
            if os.path.exists(filename):
                os.remove(filename)
            try:
                os.removedirs(os.path.dirname(filename))
            except os.error:
                # directory not empty
                pass
        # remove page index entries
        with self.index_con as con:
            q = """
                DELETE
                FROM ts_pages
                WHERE channel = '{channel}' AND page in ({pages})
            """.format(channel=channel_id, pages=','.join(map(str, pages)))
            con.execute(q)

    def page_file(self, channel_id, page, make_dir=False):
        """
        Return the file corresponding to a timeseries page (stored in feather format).
        """
        filedir = os.path.join(self.dir, filter_id(channel_id))
        if make_dir and not os.path.exists(filedir):
            os.makedirs(filedir)
        filename = os.path.join(filedir,'page-{}.feather'.format(page))
        return filename

    def clear(self):
        import shutil
        shutil.rmtree(self.dir)

    @property
    def page_files(self):
        return glob(os.path.join(self.dir,'*','*.feather'))

    @property
    def size(self):
        """
        Returns the size of the cache in bytes
        """
        all_files = self.page_files + [self.index_loc]
        return sum(map(lambda x: os.stat(x).st_size, all_files))

cache = Cache() if settings.use_cache else None



# -*- coding: utf-8 -*-

import io
import os
import sys
import uuid
import time
import boto3
import threading
from botocore.client import Config
from boto3.s3.transfer import S3Transfer
from concurrent.futures import ThreadPoolExecutor, as_completed

# blackfynn
from blackfynn import settings
from blackfynn.api.base import APIBase
from blackfynn.models import TimeSeries, Collection, Dataset
from blackfynn.utils import log

# GLOBAL
UPLOADS = {}

def check_files(files):
    for f in files:
        if not os.path.exists(f):
            raise Exception("File not found.")


class ProgressPercentage(object):
    def __init__(self, filename, upload_session_id):
        self._filename = filename
        self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()
        self._progress = 0
        self._done = False
        self._errored = False

        if upload_session_id is None:
            upload_session_id = str(uuid.uuid4())
        self._session_id = upload_session_id

    def __call__(self, bytes_amount):
        global UPLOADS
        # To simplify we'll assume this is hooked up
        # to a single filename.
        if self._errored:
            return
        with self._lock:
            self._seen_so_far += bytes_amount
            self._done = self._seen_so_far >= self._size
            UPLOADS[self._session_id] = self

    @property
    def progress(self):
        return (self._seen_so_far / self._size)

    def set_error(self):
        self._errored = True


class UploadManager(object):
    def __init__(self):
        self.id = str(uuid.uuid4())
        # map of filename -> progress callback
        self._uploads = {}
        self._exit = False

    def init_file(self, filename):
        global UPLOADS
        # filename -> session
        uid = str(uuid.uuid4())
        self._uploads[filename] = uid
        # init to blank progress
        UPLOADS[uid] = ProgressPercentage(filename, None)
        return uid

    @property
    def uploads(self):
        return [(f, UPLOADS[uid]) for f, uid in self._uploads.items()]

    @property
    def done(self):
        return sum([fstat._done for _,fstat in self.uploads]) == len(self._uploads)

    def _print_progress(self, width=16):
        for f, fstat in self.uploads:

            if fstat._done:
                state = 'DONE'
            elif fstat._errored:
                state = 'ERRORED'
                self._exit = True
            elif fstat._seen_so_far>0:
                state = 'UPLOADING'
            else:
                state = 'WAITING'

            text = ' [ {bars}{dashes} ] {state:12s} {percent:05.1f}% {name}\n'.format(
                        bars = '#'*int(fstat.progress*width),
                        dashes = '-'*(width - int(fstat.progress*width)),
                        percent = fstat.progress*100,
                        name = os.path.basename(f),
                        state = state)

            sys.stdout.write('{}\r'.format(text))
            sys.stdout.flush()

        # move cursor to relative beginning
        sys.stdout.write("\033[F"*len(self.uploads))

    def display_progress(self):
        while True:
            self._print_progress()
            time.sleep(0.2)
            if self.done or self._exit: break
        self._print_progress()
        # move cursor to bottom
        sys.stdout.write('\n'*len(self.uploads))


def upload_file(
        file,
        s3_bucket,
        s3_keybase,
        region,
        access_key_id,
        secret_access_key,
        session_token,
        encryption_key_id,
        upload_session_id=None):

    # progress callback
    progress = ProgressPercentage(file, upload_session_id)
    UPLOADS[upload_session_id] = progress

    try:
        # account for dev connections
        resource_args = {}
        config_args = dict(signature_version='s3v4')
        if 'amazon' not in settings.s3_host.lower() and len(settings.s3_host)!=0:
            resource_args = dict(endpoint_url="http://{}:{}".format(settings.s3_host, settings.s3_port))
            config_args = dict(s3=dict(addressing_style='path'))

        # connect to s3
        session = boto3.session.Session()
        s3 = session.client('s3',
            region_name = region,
            aws_access_key_id = access_key_id,
            aws_secret_access_key = secret_access_key,
            aws_session_token = session_token,
            config = Config(**config_args),
            **resource_args
        )

        # s3 key
        s3_key = '{}/{}'.format(s3_keybase, os.path.basename(file))

        # upload file to s3
        s3.upload_file(
            Filename=file,
            Bucket=s3_bucket,
            Key=s3_key,
            Callback=progress,
            ExtraArgs=dict(
                ServerSideEncryption="aws:kms",
                SSEKMSKeyId=encryption_key_id
            ))

        return s3_key

    except Exception as e:
        log.debug(e)
        progress.set_error()


class IOAPI(APIBase):
    """
    Input/Output interface.
    """
    name = 'io'

    def upload_files(self, destination, files, append=False, display_progress=False):
        group_id = str(uuid.uuid4())
        destination_id = self._get_id(destination)

        # check input files
        check_files(files)

        # sanity check destination
        pkg = self.session.core.get(destination_id)
        if append:
            if not isinstance(pkg, TimeSeries):
                raise Exception("Append destination must be TimeSeries package.")
        else:
            if not isinstance(pkg, (Collection, Dataset)):
                raise Exception("Upload destination must be Collection or Dataset.")

        # get upload credentials
        resp = self.session.security.get_upload_credentials(group_id, destination_id)
        creds = resp['tempCredentials']
        s3_bucket = resp['s3Bucket']
        s3_keybase = '{}/{}'.format(resp['s3Key'], group_id)
        region = creds['region']
        access_key_id = creds['accessKey']
        secret_access_key = creds['secretKey']
        session_token = creds['sessionToken']
        encryption_key_id = resp['encryptionKeyId']

        # upload session manager
        upload_session = UploadManager()

        # parallel upload to s3
        results = {}
        with ThreadPoolExecutor(max_workers=min(len(files),settings.max_upload_workers)) as e:
            futures = {
                e.submit(
                    fn = upload_file,
                    file = file,
                    s3_bucket = s3_bucket,
                    s3_keybase = s3_keybase,
                    region = region,
                    access_key_id = access_key_id,
                    secret_access_key = secret_access_key,
                    session_token = session_token,
                    encryption_key_id = encryption_key_id,
                    upload_session_id = upload_session.init_file(file)
                ): file
                for file in files
            }
            # thread for displaying progress
            if display_progress:
                e.submit(upload_session.display_progress())
            # retrieve results
            for future in as_completed(futures):
                fname = futures[future]
                results[fname] = future.result()
                if future.exception() is not None:
                    raise future.exception()

        # trigger ETL import
        return self.set_upload_complete(group_id, destination_id, append)

    def set_upload_complete(self, group_id, destination_id, append=False):
        params = {} if not append else dict(append=True)
        return self._post(
            self._uri('/files/upload/complete/{group_id}/{destination_id}',
                group_id = group_id,
                destination_id = destination_id),
            params = params
            )



'''
usage:
  bf append [options] <destination> <file>...

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt

from cli_utils import get_client

def main():
    args = docopt(__doc__)

    bf = get_client()
    
    files = args['<file>']
    dest  = args['<destination>']
    bf._api.io.upload_files(dest, files, append=True, display_progress=True)



'''
usage:
  bf search [options] <term>...

search options:
  --show-paths

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
from cli_utils import print_path_tree, get_client

def main():
    args = docopt(__doc__)

    bf = get_client()

    terms = ' '.join(args['<term>'])
    results = bf._api.search.query(terms)
    if len(results) == 0:
        print "No Results."
    else:
        if args['--show-paths']:
            print_path_tree(bf, results)
        else:
            for r in results:
                print " * {}".format(r)



'''
usage:
  bf delete [options] <item>

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
from blackfynn import Dataset
from blackfynn.models import BaseNode

from cli_utils import get_client

def main():
    args = docopt(__doc__)

    bf = get_client()
    
    item = bf.get(args['<item>']) 
    if isinstance(item, Dataset):
        print "Error: cannot delete dataset"
        return
    elif not isinstance(item, BaseNode):
        print "Error: cannot delete item"
        return
    bf.delete(item)



import os
import pytest

from blackfynn.models import DataPackage, TimeSeries

def test_upload(client, dataset):
    """
    Note: ETL will fail since destination will likely be removed
          before being processed. 
    """
    srcdir = os.path.dirname(__file__)
    file1 = os.path.join(srcdir, 'test-upload.txt')
    files = [file1]

    # upload a file into dataset
    r = client.upload(dataset, *files)
    assert len(r['files']) == len(files)
    assert r['destination'] == dataset.id
    assert r['appendToPackage'] == False

    # let's do it using model method
    r = dataset.upload_files(*files)
    assert len(r['files']) == len(files)
    assert r['destination'] == dataset.id
    assert r['appendToPackage'] == False

    # try uploading into a DataPackage
    pkg = DataPackage('Rando Thing', package_type='MRI')
    dataset.add(pkg)
    assert pkg.exists

    # should definitely raise an error
    with pytest.raises(Exception):
        client.upload(pkg, *files)


def test_append(client, dataset):
    """
    Note: ETL will fail for append, because... it's a text file.
          But also because the destination will likely be removed
          before the ETL actually places the new node there.
    """
    srcdir = os.path.dirname(__file__)
    file1 = os.path.join(srcdir, 'test-upload.txt')

    # TimeSeries package to append into...
    pkg = TimeSeries('Rando Timeseries')
    dataset.add(pkg)
    assert pkg.exists

    # upload/append file into package
    r = pkg.append_files(file1)
    assert len(r['files']) == 1
    assert r['destination'] == pkg.id
    assert r['appendToPackage'] == True



'''
usage:
  bf create [options] (dataset | ds)   <name>
  bf create [options] (collection | c) <name> <destination>

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''
from docopt import docopt
from blackfynn import Collection

from cli_utils import get_client

def main():
    args = docopt(__doc__)

    bf = get_client()
    
    if args['collection'] or args['c']:
        dest = args['<destination>']
        name = args['<name>']
        c = Collection(name)
        parent = bf.get(dest)
        parent.add(c)
        print(c)
    elif args['dataset'] or args['ds']:
        name = args['<name>']
        ds = bf.create_dataset(name)
        print ds



# -*- coding: utf-8 -*-
 
import json
import base64
import logging
from blackfynn import settings
import requests
from concurrent.futures import TimeoutError
from requests_futures.sessions import FuturesSession

# blackfynn
from blackfynn.utils import log
from blackfynn.models import User


class UnauthorizedException(Exception):
    pass


class BlackfynnRequest(object):
    def __init__(self, func, uri, *args, **kwargs):
        self._func = func
        self._uri = uri
        self._args = args
        self._kwargs = kwargs
        self._request = None
        self.call()

    def _handle_response(self, sess, resp):
        log.debug("resp = {}".format(resp))
        log.debug("resp.content = {}".format(resp.content))
        if resp.status_code in [requests.codes.forbidden, requests.codes.unauthorized]:
            raise UnauthorizedException()

        if not resp.status_code in [requests.codes.ok, requests.codes.created]:
            msg = "status: {} method: {} url:{} responseBody: {}".format(
                resp.status_code, resp.request.method, resp.url, resp.content)
            log.error(msg)
            resp.raise_for_status()
        try:
            # return object from json
            resp.data = json.loads(resp.content)
        except:
            # if not json, still return response content
            resp.data = resp.content

    def call(self):
        self._request = self._func(self._uri, background_callback=self._handle_response, *self._args, **self._kwargs)
        return self

    def result(self,*args, **kwargs):
        return self._request.result(*args, **kwargs)


class ClientSession(object):
    def __init__(self, api_token=None, api_secret=None, host=None, streaming_host=None):
        self._host = host
        self._streaming_host = streaming_host
        self._api_token = api_token
        self._api_secret = api_secret

        self._session = None
        self._token = None
        self._secret = None
        self._context = None
        self.profile = None

    def authenticate(self):
        """
        API token is used to authentiate against the Blackfynn platform. An API
        session token is returned from the API call and is used for all subsequent
        API calls.
        """
        # make authentication request
        session_response = self._post('/account/api/session', data=dict(tokenId = self._api_token, secret = self._api_secret))

        # parse response, set session
        self.token = session_response['session_token']
        self.profile = User.from_dict(self._get('/user/'))

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, value):
        self._token = value
        self.session.headers.update({'X-SESSION-ID': value})

    @property
    def session(self):
        """
        Make requests-futures work within threaded/distributed environment.
        """
        if not hasattr(self._session, 'session'):
            self._session = FuturesSession(max_workers=4)
            self._session.headers.update({'X-SESSION-ID': self._token})

        return self._session

    def _make_call(self, func, uri, *args, **kwargs):
        log.debug("uri = {}".format(uri))
        log.debug("args = {}".format(args))
        log.debug("kwargs = {}".format(kwargs))
        return BlackfynnRequest(func, uri, *args, **kwargs)

    def _call(self, method, endpoint, base='', async=False, *args, **kwargs):
        if method == 'get':
            func = self.session.get
        elif method == 'put':
            func = self.session.put
        elif method == 'post':
            func = self.session.post
        elif method == 'delete':
            func = self.session.delete

        # serialize data
        if 'data' in kwargs:
            kwargs['data'] = json.dumps(kwargs['data'])

        # we might specify a different host
        if 'host' in kwargs:
            host = kwargs['host']
            kwargs.pop('host')
        else:
            host = self._host

        # call endpoint
        uri = self._uri(endpoint, base=base, host=host)
        req = self._make_call(func, uri, *args, **kwargs)

        if async:
            return req
        else:
            return self._get_response(req)

    def _uri(self, endpoint, base, host=None):
        if host is None:
            host = self._host
        return '{}{}{}'.format(host, base, endpoint)

    def _get(self, endpoint, async=False, *args, **kwargs):
        return self._call('get', endpoint, async=async, *args, **kwargs)

    def _post(self, endpoint, async=False, *args, **kwargs):
        return self._call('post', endpoint, async=async, *args, **kwargs)

    def _put(self, endpoint, async=False, *args, **kwargs):
        return self._call('put', endpoint, async=async, *args, **kwargs)

    def _del(self, endpoint, async=False, *args, **kwargs):
        return self._call('delete', endpoint, async=async, *args, **kwargs)

    def _get_result(self, req, count=0):
        try:
            resp = req.result(timeout=settings.max_request_time)
        except TimeoutError as e:
            if count < settings.max_request_timeout_retries:
                # timeout! trying again...
                resp = self._get_result(req.call(), count=count+1)
        except UnauthorizedException as e:
            # try refreshing the session
            if self._token is not None and count==0:
                self.authenticate()
                # re-request
                resp = self._get_result(req.call(), count=count+1)
            else:
                raise e
        return resp

    def _get_response(self, req):
        resp = self._get_result(req)
        return resp.data

    def register(self, *components):
        """
        Register API component with session. Components should all be of 
        APIBase type and have a name and base_uri property. 

        The registered component will have reference to base session to 
        make higher-level calls outside of its own scope, if needed.
        """
        # initialize
        for component in components:
            c = component(session=self)
            assert len(component.name) > 1, "Invalid API component name"
            # component is accessible via session.(name)
            self.__dict__.update({ component.name: c })

    @property
    def headers(self):
        return self.session.headers



import os
import pytest
import requests
import datetime

# client library
from blackfynn.models import (
    DataPackage, Dataset, Collection, File
)
from blackfynn.base import UnauthorizedException

def test_get_all_datasets(client, superuser_client):
    n_ds_to = len(client.datasets())

    ds1 = client.create_dataset("DS 1")
    ds2 = client.create_dataset("DS 2")

    assert len(client.datasets()) == n_ds_to + 2

    ds3 = client.create_dataset("DS 3")
    assert len(client.datasets()) == n_ds_to + 3

    # cleanup
    superuser_client.delete(ds1)
    superuser_client.delete(ds2)
    superuser_client.delete(ds3)

    assert len(client.datasets()) == n_ds_to


def test_datasets(dataset):
    ds_items = len(dataset)

    # create package locally
    pkg = DataPackage('Child of Dataset', package_type='Text')

    assert not pkg.exists
    assert pkg not in dataset

    # add package to dataset
    dataset.add(pkg)

    assert pkg.exists
    assert pkg in dataset
    assert pkg.id in dataset
    assert len(dataset) == ds_items + 1

    # remove from dataset
    dataset.remove(pkg)

    assert not pkg.exists
    assert pkg not in dataset
    assert len(dataset) == ds_items

# def test_dataset_collaborators(client, client2, test_organization, dataset):
#     # share with user
#     user2Id = client2.profile.id
#     user_share_resp = dataset.add_collaborators(user2Id)
#     assert user_share_resp['changes'][user2Id]['success']

#     # unshare with user
#     user_remove_resp = dataset.remove_collaborators(user2Id)
#     assert user_remove_resp['changes'][user2Id]['success']

#     # # share with org
#     org_share_resp = dataset.add_collaborators(test_organization.id)
#     assert org_share_resp['changes'][test_organization.id]['success']

#     # # unshare with org
#     org_remove_resp = dataset.remove_collaborators(test_organization.id)
#     assert org_remove_resp['changes'][test_organization.id]['success']

def test_packages_create_delete(client, dataset):

    # init
    pkg = DataPackage('Some MRI', package_type='MRI')
    assert not pkg.exists

    # create
    dataset.add(pkg)
    assert pkg.exists
    assert pkg.id is not None
    assert pkg.name == 'Some MRI'

    # TODO: (once we auto-include in parent)
    assert pkg in dataset

    # update package name
    pkg.name = 'Some Other MRI'
    pkg = client.update(pkg)

    pkg2 = client.get(pkg.id)

    assert pkg2.name == 'Some Other MRI'
    assert pkg2.id == pkg.id

    # delete all packages
    client.delete(pkg)

    assert not pkg.exists

    pkg = DataPackage('Something else', package_type='TimeSeries')
    assert not pkg.exists
    dataset.add(pkg)
    assert pkg.exists
    pid = pkg.id
    pkg.delete()
    assert not pkg.exists

    with pytest.raises(Exception):
        pkg2 = client.get(pid)

    # TODO: (once we auto-remove from parent)
    #assert pkg not in dataset

def test_package_states(client, dataset):
    pkg = DataPackage('My Stateful Package', package_type='Slide')
    assert not pkg.exists
    dataset.add(pkg)
    assert pkg.exists
    assert pkg.state == "UNAVAILABLE"
    pkg.set_ready()
    pkg2 = client.get(pkg.id)
    assert pkg2.id == pkg.id
    assert pkg2.state == pkg.state
    assert pkg2.state == "READY"
    pkg.set_error()
    del pkg2

    pkg2 = client.get(pkg.id)
    assert pkg2.id == pkg.id
    assert pkg2.state == pkg.state
    assert pkg2.state == "ERROR"
    del pkg2

    pkg.delete()


def test_properties(client, dataset):

    pkg = DataPackage('Some Video', package_type='Video', parent=dataset)
    assert not pkg.exists

    dataset.add(pkg)
    assert pkg.exists

    pkg.insert_property('my-key','my-value')
    pkg2 = client.get(pkg)
    print 'properties =', pkg2.properties
    assert pkg2.id == pkg.id
    assert pkg2.get_property('my-key').data_type == 'string'
    assert pkg2.get_property('my-key').value == 'my-value'

    explicit_ptypes = {
        'my-int1': ('integer', 123123),
        'my-int2': ('integer', '123123'),
        'my-float': ('double', 123.123),
        'my-float2': ('double', '123.123'),
        'my-float3': ('double', '123123'),
        'my-date': ('date', 1488847449697),
        'my-date2': ('date', 1488847449697.123),
        'my-date3': ('date', datetime.datetime.now()),
        'my-string': ('string', 'my-123123'),
        'my-string2': ('string', '123123'),
        'my-string3': ('string', '123123.123'),
        'my-string4': ('string', 'According to plants, humans are blurry.'),
    }
    for key, (ptype,val) in explicit_ptypes.items():
        pkg.insert_property(key, val, data_type=ptype)
        assert pkg.get_property(key).data_type == ptype

    inferred_ptypes = {
        'my-int1': ('integer', 123123),
        'my-int2': ('integer', '123123'),
        'my-float1': ('double', 123.123),
        'my-float2': ('double', '123.123'),
        'my-date': ('date', datetime.datetime.now()),
        'my-string': ('string', 'i123123'),
        'my-string2': ('string', '#1231'),
    }
    for key, (ptype,val) in inferred_ptypes.items():
        pkg.insert_property(key, val)
        prop = pkg.get_property(key)
        assert prop.data_type == ptype

    # remove property
    pkg.remove_property('my-key')
    with pytest.raises(Exception):
        assert pkg.get_property('my-key')

    pkg2 = client.get(pkg.id)
    with pytest.raises(Exception):
        assert pkg2.get_property('my-key')

def test_package_objects(client, superuser_client, dataset):
    """
    Only super-users are allowed to create/modify package sources/files.
    """ 
    pkg = DataPackage('Some Video', package_type='Video')
    assert not pkg.exists

    # some files (local for now)
    source = File(name='My Source File', s3_key='s3/source', s3_bucket='my-bucket', file_type="JSON")
    file   = File(name='My File', s3_key='s3/file', s3_bucket='my-bucket', file_type="CSV")
    view   = File(name='My View File', s3_key='s3/view', s3_bucket='my-bucket', file_type="NIFTI")

    # get dataset (but as super-user)
    superuser_dataset = superuser_client.get(dataset.id)
    assert superuser_dataset.id == dataset.id
    assert superuser_dataset.exists
    assert superuser_dataset.type == dataset.type

    # create package (super-admin user session)
    superuser_dataset.add(pkg)
    assert pkg.exists

    # add source (super-admin)
    pkg.set_sources(source)

    # get as normal user
    pkg2 = client.get(pkg) 
    print "sources =", pkg2.sources
    assert len(pkg2.sources) > 0
    assert pkg2.sources[0].name == 'My Source File'
    del pkg2

    # add files (super-admin)
    pkg.set_files(file)

    # get as normal user
    pkg2 = client.get(pkg)
    print "files =", pkg2.files
    assert len(pkg2.files) > 0
    del pkg2

    # add views (super-admin)
    pkg.set_view(view)

    # get as normal user
    pkg2 = client.get(pkg)
    print "view =", pkg2.view
    assert len(pkg2.view) > 0
    del pkg2
    del pkg

    # create package (normal user owns)
    pkg = DataPackage('Some Video', package_type='Video', parent=dataset)
    assert not pkg.exists
    dataset.add(pkg)
    assert pkg.exists

    # try doing as normal user - should error
    with pytest.raises(UnauthorizedException):
        pkg.set_sources(source)

    with pytest.raises(UnauthorizedException):
        pkg.set_files(file)

    with pytest.raises(UnauthorizedException):
        pkg.set_view(view)

        



'''
usage:
  bf dataset [options] <dataset> [<command>] [<action>] [<action-args>...]

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''

from docopt import docopt
import sys

from cli_utils import get_client

def main():
    args = docopt(__doc__)

    bf = get_client()
    
    ds = bf.get(args['<dataset>'])
    if args['collaborators']:
        if args['<action>'] == 'ls':
            resp = ds.collaborators()
            print " - Users"
            for u in resp['users']:
                print "   - email:{} id:{}".format(u.email, u.id)
            print " - Groups"
            for g in resp['groups']:
                print "   - name:{} id:{}".format(g.name, g.id)
        elif args['<action>'] == 'add':
            ids = args['<action-args>']
            if len(ids) == 0:
                print "Error: No ids specified"
                sys.exit(1)
            resp = ds.add_collaborators(*ids)
            print_collaborator_edit_resp(resp)
        elif args['<action>'] == 'rm':
            ids = args['<action-args>']
            if len(ids) == 0:
                print "Error: No ids specified"
                sys.exit(1)
            resp = ds.remove_collaborators(*ids)
            print_collaborator_edit_resp(resp)
        else:
            print "Error: invalid dataset collaborators command. Valid commands are 'ls', 'add' or 'rm'"

    
